/*
'******************************************************************************
' FileName: Compile.cs
' Author:   VL, HS, RS
' Created:  Nov-2003
' This is the main entrypoint into the PWHCOMP tool. 
' Pwhcomp has the responsibility of taking a PWH page and creating
' code for it such that it is executable and calls the codebehind accordingly.
'*****************************************************************************
*/

using System;
using System.IO;
using System.Text;
using System.Collections;
using System.Collections.Specialized;

namespace pwhcomp
{	
	// Below are data types used during the compilation phases.
	enum BlockType: long
	{
		HadCode,
		HasNoCode
	}

	enum TagControlType: long
	{
		InputBox,
		TextBox
	}

	struct ParseBlock
	{
		public string start_decorator;
		public string end_decorator;
	}

	struct ContentBlock
	{
		
		public BlockType ContentType;
		public int start;
		public int code_start;
		public int end;
		public int code_end;
		//public string content;
	}

	class Compile
	{	
	
		public class Code
		{
			public string ConstructTextCodeBlock(string mQueryString)
			{
				StringBuilder sReturnQueryString = new StringBuilder("");

				if(mQueryString.Length > 0)
				{
     
					for (int i = 0; i < mQueryString.Length; i++)
					{
						if (mQueryString[i].Equals('"'))
						{
							sReturnQueryString.Append(mQueryString[i]);
							sReturnQueryString.Append("\"");
						}
						else
						{
							sReturnQueryString.Append(mQueryString[i]);
						}
					}
				}
				return "Response.Write(@\"" + sReturnQueryString.ToString() + "\");";
			}

			//ConstructCodeBlock
			//This method will Wrap the textual components into a Response.Wrire block
			//Such that it will render as regular HTML
			public string ConstructCodeBlock(string block)
			{
		
				String fixblock = block.Replace(Convert.ToString('\"'),"\"\"");
				
				String ntext = "Response.Write(@\"" + fixblock + "\");"; 
				return ntext;
			}

			public string ConstructGeneric(Queue q, string fs)
			{
				// Load in the page template, 
				// Make note of the PWH: tag callouts
				// Load the template into the file.
				string bigfile = "";

				try
				{

					string yu_text;
					StreamReader sr = new StreamReader("template.cs");
					string temp = sr.ReadToEnd();					
					

					string newtext = "";
					string cd_txt = "";
					string controlbuilder_txt = "";

					ContentBlock yu = new ContentBlock();

				
					int cto = q.Count;
					//Lets develop the start code that will be inserted.
					for (int i = 0;i<cto;i++)
					{
						yu = (ContentBlock)q.Dequeue();
						if (yu.ContentType == BlockType.HadCode)
						{
							//treat all of the stuff before this as normal.
							yu_text = fs.Substring(yu.start,yu.end-yu.start);
							newtext = newtext + ConstructTextCodeBlock(yu_text);
							// Now lets focus on the contents of the control that was given.
							cd_txt = fs.Substring(yu.code_start,(yu.code_end-yu.code_start));
							controlbuilder_txt = controlbuilder_txt + HandleControlBuilderTag(cd_txt);
							cd_txt = HandlePWHTag(cd_txt);
							newtext = newtext + cd_txt;

							
						}
						else
						{
							yu_text = fs.Substring(yu.start,yu.end);
							newtext = newtext + ConstructTextCodeBlock(yu_text);
						}
					}
					
					string s_all="";

					string end_of_controls = "string path = System.IO.Path.GetDirectoryName(System.Reflection.Assembly.GetExecutingAssembly().GetName().CodeBase );\n";
					end_of_controls= end_of_controls + "string assemblyname = path + \"\\\\%%PWH_CB_CLASS%%.dll\";\n";
					end_of_controls= end_of_controls + "Assembly a = Assembly.LoadFrom(assemblyname);\n";
					end_of_controls= end_of_controls + "Type page = a.GetType(\"%%PWH_NAMESPACE%%.%%PWH_CB_CLASS%%\");\n";
					end_of_controls= end_of_controls + "object obj = Activator.CreateInstance(page);\n";
					end_of_controls= end_of_controls + "MethodInfo oninit = page.GetMethod(\"OnInit\");\n";
					end_of_controls= end_of_controls + "MethodInfo pageLoad = page.GetMethod(\"Page_Load\");\n";
					end_of_controls= end_of_controls + "object[] paramArray = new object[1];\n";
					end_of_controls= end_of_controls + "paramArray[0] = Response.GetCtlContainer();\n";
					end_of_controls= end_of_controls + "object ret_obj = oninit.Invoke(obj, paramArray);\n";
					end_of_controls= end_of_controls + "ret_obj = pageLoad.Invoke(obj, null);\n";

										
					
					s_all = controlbuilder_txt + end_of_controls + newtext;

					temp = temp.Replace("%%START%%",s_all);						
					
					bigfile = temp;
					// Now replase te start code with the code blocks.


				}
				catch (Exception e) 
				{
					// Let the user know what went wrong.
					Console.WriteLine("The file could not be read:");
					Console.WriteLine(e.Message);
				}


				return bigfile;
			}

			//HandleControlBuilderTag
			//
			public string HandleControlBuilderTag(string cd_txt)
			{
				// Instantiate the pwhcomp's tag class
				comp_control control_lib = new comp_control();
				string newtext = control_lib.HandleControlBuilderTag(cd_txt);
				
				return newtext;
			}

			// HandlePWHTag
			// This method will call the compile time ASP tag comp_control class
			public string HandlePWHTag(string cd_txt)
			{
				// Instantiate the pwhcomp's tag class
				comp_control control_lib = new comp_control();
				string newtext = control_lib.compiler_generatecode(cd_txt);
				
				return newtext;
			}
		}
		
		[STAThread]
			//Below is the main entry point in the execute of this code.
		static void Main(string[] args)
		{
	
			StreamReader sr = new StreamReader(args[0]);
		    
			
			string out_assembly = "";
			int suffix_location = args[0].IndexOf(".pwh");
			string req_assembly = args[0].Substring(0,suffix_location);
			out_assembly = "CB_EX_" + req_assembly;  // This will be inserted in the template as %%PWH_PAGE%%
			string test = sr.ReadToEnd();	
			string newtest = "";
			
			string CB_Namespace=args[1];
			string CB_Class = args[2];
			
			

			int s_len = test.Length;
			bool endoffile=false;
			
			int i=0;
			int j=0;
			Code cd = new Code();
			int nonstart = 0;
			Queue myQ = new Queue();

			ParseBlock CodeBlock = new ParseBlock();
			//
			CodeBlock.start_decorator = "<PWH:";
			CodeBlock.end_decorator = "/PWH>";
			
			// In this section, we are only interested in finding the location (if any) of the ASP: Blocks
			int fp_start_byte = 0;

			while (!endoffile)
			{

				// Find instances of the ASP Block in the file supplied.
				
				i = test.IndexOf(CodeBlock.start_decorator,fp_start_byte);
				if (i!=-1)

				{   
					
				
					j = test.IndexOf(CodeBlock.end_decorator,i,test.Length-i);
					if (j!=1)
					{
						// OK Get the substring of the tag

						ContentBlock ct = new ContentBlock();
						ct.ContentType = BlockType.HadCode;
						ct.start = fp_start_byte;
						//ct.code_start = i+8;
						ct.code_start = i+5;
						ct.end = i-1;
						ct.code_end = j+8;
						myQ.Enqueue(ct);
						
						fp_start_byte = ct.code_end;
						// For each item in the queue, determine the type and strip out the blocks as needed.

					}
					else
					{
						// This is in error
					}
				}
				else
				{

					// There was no code, and therefore with gice you a nice block block of text that looks as follows:
					ContentBlock ct2 = new ContentBlock();
					ct2.ContentType = BlockType.HasNoCode;
					ct2.start = fp_start_byte;
					ct2.code_start = 0;
					ct2.code_end = 0;
					ct2.end = test.Length-fp_start_byte;

					myQ.Enqueue(ct2);	
					endoffile = true;
				}
				
			}
			
			test = cd.ConstructGeneric(myQ, test);
			test = test.Replace("%%PWH_PAGE%%",req_assembly);
			test = test.Replace("%%PWH_NAMESPACE%%",CB_Namespace);
			test = test.Replace("%%PWH_CB_CLASS%%",CB_Class);			

			StreamWriter srx = File.CreateText(out_assembly+ ".cs");
			srx.Write(test);
					
			srx.Close();

		}
	}
				
}

