/*
'******************************************************************************
' FileName: WorkerThread.cs
' Author:   VL, HS, RS
' Created:  Nov-2003
' This class handles the actual connections that were made to the the Listening port.
' Once a connection is made, one of these treads handles the incoming request.
' This module handles HTML pages, Images, and PWH page processing.
'*****************************************************************************
*/

 
using System;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Diagnostics;
using System.Text;
using System.Reflection;

namespace pwhost
{

	public class WorkerThread
	{

		public static string HttpdVersion = "pwhost 0.3";

		private TcpClient client;
		private byte[] recByte = new byte[1048576];
		private string physicalDirectory;
		private AsyncCallback GetMsgCallback;
		private MemoryStream memStream;

		public WorkerThread(TcpClient client, string physicalDirectory)
		{
			memStream = new MemoryStream();
			Debug.WriteLine("Client class constructor reached.");
			this.physicalDirectory = physicalDirectory;
			this.client = client;
			GetMsgCallback = new AsyncCallback(GetMessage);
			Run();
		}

		internal void Run() 
		{
				(client.GetStream()).BeginRead(recByte,0,1048576,GetMsgCallback,null);
		}

		//
		//GetMessage
		//----------
		//This method will obtain the query made to the thread and send it to the ProcessMessage.		
		public void GetMessage(IAsyncResult ar)
		{
			int byteCount;
			try
			{
				//Get the number of Bytes received
				byteCount = (client.GetStream()).EndRead(ar);
				if(byteCount <1)
				{
					//Close the socket
					client.Close();
					return;
				}

				System.Text.StringBuilder myString = new System.Text.StringBuilder();

				for(int i=0; i<byteCount; i++)
				{	
					myString.Append(Convert.ToChar(recByte[i]));
					memStream.WriteByte(recByte[i]);
				}

				// Check to see if we have a Content-Length and \r\n\r\n then get all the data, 
				// or no content length but a \r\n\r\n
				int contentLength = getContentLength(System.Text.Encoding.ASCII.GetString(memStream.ToArray(),0,Convert.ToInt32(memStream.Length)));
				string memStreamString = System.Text.Encoding.ASCII.GetString(memStream.ToArray(),0,Convert.ToInt32(memStream.Length));
				
				if ((contentLength == 0) && (memStreamString.IndexOf("\r\n\r\n")>0))
				{
					memStream = new MemoryStream();
					ProcessMessage(memStreamString);
					recByte = new Byte[1048576];
				}
				else if (contentLength>0) 
				{
					memStream = new MemoryStream();
					ProcessMessage(memStreamString);
					recByte = new Byte[1048576];
				}
	
				(client.GetStream()).BeginRead(recByte,0,1048576,GetMsgCallback,null);

			}
			catch (Exception ex)
			{
				client.Close();
				
			}
		}
	
		//ProcessMessage
		//--------------
		//GetMessage passes the actual query to this method.  This is among one of the more interesting methods in the 
		//Class.  This section also is also enlightening in terms of the HTTP protocol.
		public void ProcessMessage(string buffer)
		{
			try
			{
				string request;
				string dirname;
				string requestedfile;
				int startposition = 0;				
				string querystring = "";
				string poststring = "";
				int getOrPostLength = 3;
				bool isCGI = false;
				bool isNotApp = false;

				
				if (buffer.Substring(0,4) == "POST" )
				{
					querystring = QueryStringProcess(buffer);			
					querystring = ProcessPost(buffer);
					buffer = QueryStringCut(buffer);					
					
					getOrPostLength = 4;
				}
				if (buffer.Substring(0,3) == "GET")
				{
					querystring = QueryStringProcess(buffer);
					buffer = QueryStringCut(buffer);
					getOrPostLength = 3;
				}
			

				startposition = buffer.IndexOf("HTTP",1);				
				string sHttpVersion = buffer.Substring(startposition,8);
			
				// now see what request it's looking for
				request = buffer.Substring(0,startposition - 1);

				//Replace backslash with Forward Slash, if Any
				request.Replace("\\","/");

				//If file name is not supplied add forward slash to indicate 
				//that it is a directory and then we will look for the 
				//default file name..
				if ((request.IndexOf(".") <1) && (!request.EndsWith("/")))
				{
					request = request + "/"; 
				}

				//Extract the requested file name
				startposition = request.LastIndexOf("/") + 1;
				requestedfile = request.Substring(startposition);
				if(requestedfile=="")
				{
					requestedfile = "default.htm";
				}
				
				// In this section, we attempt to determine if the requested resource is an ASPX page
				// or not.  We bool this with the notion that it is or is not an application.
				if (requestedfile.EndsWith(".pwh"))
				{
					isNotApp = false;
				}
				else
				{
					isNotApp = true;
				}
			
				dirname = request.Substring(request.IndexOf("/"), request.LastIndexOf("/")-getOrPostLength);
				dirname = dirname.Replace("/","\\");

				string path;
				path = System.IO.Path.GetDirectoryName( 
					System.Reflection.Assembly.GetExecutingAssembly().GetName().CodeBase );
				//path = path.Remove(0,6);
				// .NET CF needs to have the whole path, the Win32 version does not and can work relative.
				// Uncomment the following line and comment out th eline below to allow Windows CE to work
				//if (!File.Exists(path + "\\" + requestedfile))
				//if (!File.Exists(requestedfile))
				//{
				//	SendToBrowser(sHttpVersion,"",1,"404",System.Text.ASCIIEncoding.ASCII.GetBytes("......"));
				//	return;
				//}


				// create the resource object
				Resource res = new Resource(physicalDirectory + dirname, requestedfile);
				res.setQueryData(querystring);

		
				// NOTE - put in CGI Support at some point
				if (isCGI)
				{
					// No Implementation.
				}
				else
				{

					if (isNotApp) 
					{					
						FileStream fs;
						if (dirname.Equals(""))
						{
							fs = new FileStream(path + "\\" + requestedfile, FileMode.Open, FileAccess.Read);
						}
						else
						{
							fs = new FileStream(path  + dirname  + requestedfile, FileMode.Open, FileAccess.Read);
						}
						BinaryReader r = new BinaryReader(fs);
						Byte[] result = r.ReadBytes(Convert.ToInt32(fs.Length));
						//
						// send the result of the process to the Client
						//
						SendToBrowser(sHttpVersion,res.getMimeType(),result.Length,"200 OK",result);

					}
					else
					{
						// ASPX Page Processing - The best part.
						// Load the requested files type, and invoke interface
						// We pass in the essentials - I.e. data coming from the browser
						// We wait for the result in return of the invocation.

						// Load the aspx namespace and invoke.
						// Once this is doen the flow will be
						// Run The ASPX class, which will call the code behind - output reulstant buffer
						string req_assembly = "";
						int suffix_location = requestedfile.IndexOf(".pwh");
						req_assembly = requestedfile.Substring(0,suffix_location);

						// In addition to the HTTPPage type, The assembly must be monikered by us
						// as this is not usually touched except by the pwhcomp tool.
						string aspx_assemblyname = path + "\\" + dirname.Remove(0,1) + "CB_EX_" + req_assembly + ".dll";						
						string type_aspx_assemblyname = "CB_EX_" + req_assembly;
						Assembly a = null;
						try
						{
							a = Assembly.LoadFrom(aspx_assemblyname);
						}
						catch(FileNotFoundException e)
						{

						}

						// Get the HTTPPage type.
						try

						{						
							Type aspx_page = a.GetType(type_aspx_assemblyname + ".HTTPPage");							
							// This type is required, pwhost needs 
							// to Create the HTTPPage type and invoke the method to get things started.
							object obj = Activator.CreateInstance(aspx_page);
							// Get info for HTTPPage.Init.
							MethodInfo aspx_init = aspx_page.GetMethod("Init");

							object[] paramArray = new object[2];
							paramArray[0] = querystring;                // INPUT  - Variables passed in
							paramArray[1] = buffer;                     // OUTPUT - Buffer contents

							// Invoke method
							object ret_obj = aspx_init.Invoke(obj, paramArray);
							
							Byte[] result = Encoding.UTF8.GetBytes(ret_obj.ToString());						

						// send the result of the process to the Client
						SendToBrowser(sHttpVersion,res.getMimeType(),result.Length,"200 OK",result);
						}
						catch (Exception e)
						{
							int i = 0;
							//Need to add some corrective measures.
						}


					}
				}
			}
			catch (Exception ex)
			{
				//Need to add some corrective measures.
			}
		}


		// 
		// gotContent
		// ----------
		// Returns bool is the message seems about right.
		//
		public bool gotContent(int contentLength, string message)
		{
			if (message.IndexOf("\r\n\r\n")>0)
			{
				int start = message.IndexOf("\r\n\r\n")+4;
				if (message.Length-contentLength==start)
					return true;
				else
					return false;
			}
			else
			{
				return false;
			}
		}

		//
		// getContentLength
		// ----------------
		// Returns the length of the content. this is a common operation
		// when dealing with HTTP
		//
		public int getContentLength(string message)
		{
			// check to see if the message has a content length
			if (!(message.IndexOf("Content-Length:")>0))
				return 0;

			StringReader sr = new StringReader(message);

			string temp = sr.ReadLine();
			while ((!(temp==null)) && !(temp.StartsWith("Content-Length:")))
			{
				temp = sr.ReadLine();
			}
			if (temp==null)
			{
				return 0;
			}
			int contentLength = int.Parse(temp.Substring(15,temp.Length-15).Trim());
			return contentLength;
		}
		
		//
		// ProcessPost
		// -----------
		// Returns the POST data received from the client.
		//
		public string ProcessPost(string post)
		{
			
			StringReader sr = new StringReader(post);

			string temp = sr.ReadLine();
			while ((!(temp==null)) && !(temp.StartsWith("Content-Length:")))
			{
				temp = sr.ReadLine();
			}
			// temp should contain content-length
			int contentLength = int.Parse(temp.Substring(15,temp.Length-15).Trim());
			int indexData = post.IndexOf("\r\n\r\n",0,post.Length);
			int postLength = post.Length;

			return post.Substring(indexData+4,contentLength);
		}

		//
		// QueryStringCut
		// --------------
		// 
		public string QueryStringCut(string query)
		{
			StringReader sr = new StringReader(query);
			string temp = sr.ReadLine();

			if (temp.IndexOf("?")>0)
			{
				string tempString = temp.Substring(0,temp.IndexOf("?"));
				int startposition = temp.IndexOf("HTTP",1);
				tempString+= " " +  temp.Substring(startposition,8);
				return tempString;
			}
			else
				return query;
		}

		//
		// QueryStringProcess
		// ------------------
		//
		public string QueryStringProcess(string query)
		{
			StringReader sr = new StringReader(query);
			string temp = sr.ReadLine();
			if (temp.IndexOf("?")>0)
			{
				string tempString = temp.Substring(query.IndexOf("?")+1,temp.IndexOf("HTTP")-temp.IndexOf("?")-1);
				return tempString.Trim();
			}
			else
				return "";
		}

		//
		// QueryPostStringReturnSingle
		// ---------------------------
		//
		public string QueryPostStringReturnSingle(string querystring, string post)
		{
			if ((post==null) || (post==""))
				return querystring;
			if ((querystring==null) || (querystring==""))
				return post;

			string[] queryarray = querystring.Split('&');
			string[] postarray = post.Split('&');
			string[] queryvar = querystring.Split('&');
			string[] postvar = post.Split('&');

			string returnString = post;
			bool found = false;
			// post over querystring
			
			for (int i=0;i<queryarray.Length;i++)
			{
				queryvar[i] = queryarray[i].Substring(0,queryarray[i].IndexOf("="));
			}
			for (int i=0;i<postarray.Length;i++)
			{
				postvar[i] = postarray[i].Substring(0,postarray[i].IndexOf("="));
			}


			for (int i=0;i<queryvar.Length;i++)
			{
				found = false;

				for (int j=0;j<postvar.Length;j++)
				{
					if (queryvar[i]==postvar[j])
					{
						found = true;
					}
				}

				if (!found)
				{
					returnString+= "&" + queryarray[i].ToString();
				}
			}
			return returnString;
		}


		//
		// SendToBrowser
		// -------------
		// This method provides a payload mechanism to allow content to be
		// delivered to the client's browser.
		public void SendToBrowser(string httpver, string mimeheader, long totalbytes, string statuscode, Byte[] buf)
		{

			try
			{
				String buffer = "";
			
				
				if (mimeheader.Length == 0 )
				{
					mimeheader = "text/html";  // Default Mime Type is text/html
				}

				buffer+= httpver + " " + statuscode + "\r\n";
				buffer+= "Server: " + HttpdVersion + "\r\n";
				buffer+= "Content-Type: " + mimeheader + "\r\n";
				buffer+= "Accept-Ranges: bytes\r\n";
				buffer+= "Content-Length: " + totalbytes + "\r\n\r\n";
			
				Byte[] sendData = System.Text.Encoding.UTF8.GetBytes(buffer); 
				

				NetworkStream stream = client.GetStream();

				stream.Write(sendData,0,sendData.Length);
				stream.Write(buf,0,buf.Length);


				if (httpver=="HTTP/1.0")
				{					
					stream.Flush();
					stream.Close();
					client.Close();
				}
			}
			catch (Exception ex)
			{

			}
		}
	
	}
}
